import argparse
import asyncio
import json
import os
import signal
import sys
import time
import traceback
from abc import ABC, abstractmethod
from argparse import ArgumentDefaultsHelpFormatter, Namespace
from asyncio import Task
from datetime import datetime, timezone
from enum import Enum, IntEnum
from importlib.metadata import EntryPoint, entry_points, version
from pathlib import Path
from secrets import token_urlsafe
from tempfile import gettempdir
from typing import Any, Optional, cast

import aiofiles
import argcomplete

from gallia.utils import camel_to_snake
from gallia.db.db_handler import DBHandler
from gallia.penlab import Dumpcap, PowerSupply, PowerSupplyURI
from gallia.penlog import Logger
from gallia.transports.base import BaseTransport, TargetURI
from gallia.transports.can import ISOTPTransport, RawCANTransport
from gallia.transports.doip import DoIPTransport
from gallia.transports.tcp import TCPLineSepTransport
from gallia.uds.ecu import ECU


class ExitCodes(IntEnum):
    SUCCESS = 0
    GENERIC_ERROR = 1
    SETUP_FAILED = 10
    TEARDOWN_FAILED = 11


class FileNames(Enum):
    PROPERTIES_PRE = "PROPERTIES_PRE.json"
    PROPERTIES_POST = "PROPERTIES_POST.json"


class Formatter(ArgumentDefaultsHelpFormatter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # This is for steff.
        help_width = os.getenv("GALLIA_HELP_WIDTH")
        if help_width is not None:
            kwargs["width"] = int(help_width)
        super().__init__(*args, **kwargs)


class GalliaBase(ABC):
    """Base class for all gallia tasks"""

    def __init__(self) -> None:
        self.artifacts_dir: Path
        self.description = self.__class__.__doc__
        self.logger = Logger(component="gallia", flush=True)
        self.db_handler: Optional[DBHandler] = None
        self.parser = argparse.ArgumentParser(
            description=self.description, formatter_class=Formatter
        )
        self.id = camel_to_snake(self.__class__.__name__)
        self._add_class_parser()
        self.add_parser()

    @abstractmethod
    def add_parser(self) -> None:
        ...

    @abstractmethod
    async def main(self, args: Namespace) -> None:
        ...

    @abstractmethod
    async def setup(self, args: Namespace) -> None:
        ...

    @abstractmethod
    async def teardown(self, args: Namespace) -> None:
        ...

    async def _run(self, args: Namespace) -> int:
        exit_code: int = ExitCodes.SUCCESS

        try:
            if args.db is not None:
                self.db_handler = DBHandler(args.db)
                await self.db_handler.connect()

                await self.db_handler.insert_run_meta(
                    script=sys.argv[0].split()[-1],
                    arguments=sys.argv[1:],
                    start_time=datetime.now(timezone.utc).astimezone(),
                    path=self.artifacts_dir,
                )

            try:
                await self.setup(args)
            except BrokenPipeError as e:
                exit_code = ExitCodes.GENERIC_ERROR
                self.logger.log_critical(str(e))
            except Exception as e:
                self.logger.log_critical(f"setup failed: {e.__class__.__name__}: {e}")
                sys.exit(ExitCodes.SETUP_FAILED)

            try:
                try:
                    await self.main(args)
                except Exception as e:
                    exit_code = ExitCodes.GENERIC_ERROR
                    self.logger.log_critical(str(e))
                    traceback.print_exc()
            finally:
                try:
                    await self.teardown(args)
                except Exception as e:
                    self.logger.log_critical(
                        f"teardown failed: {e.__class__.__name__}: {e}"
                    )
                    sys.exit(ExitCodes.TEARDOWN_FAILED)
            return exit_code
        except KeyboardInterrupt:
            exit_code = 128 + signal.SIGINT
            raise
        except SystemExit as se:
            exit_code = se.code
            raise
        finally:
            if self.db_handler is not None:
                try:
                    await self.db_handler.complete_run_meta(
                        datetime.now(timezone.utc).astimezone(), exit_code
                    )
                except Exception as e:
                    self.logger.log_warning(
                        f"Could not write the run meta to the database: {repr(e)}"
                    )

                try:
                    await self.db_handler.disconnect()
                except Exception as e:
                    self.logger.log_debug(
                        f"Could not close the database connection {repr(e)}"
                    )

    def _add_class_parser(self) -> None:
        group = self.parser.add_argument_group("generic gallia arguments")
        group.add_argument(
            "--data-dir",
            default=os.environ.get("PENRUN_ARTIFACTS"),
            type=Path,
            help="Folder for artifacts",
        )
        group.add_argument(
            "--db",
            default=os.environ.get("GALLIA_DB"),
            type=Path,
            help="Path to sqlite3 database",
        )

    def run(self) -> int:
        argcomplete.autocomplete(self.parser)
        args = self.parser.parse_args()

        if args.data_dir is None:
            base = Path(gettempdir())
            self.artifacts_dir = base.joinpath(
                f'{self.id}_{time.strftime("%Y%m%d-%H%M%S")}_{token_urlsafe(6)}'
            )
            self.artifacts_dir.mkdir(parents=True)
        else:
            if not args.data_dir.is_dir():
                self.logger.log_error(
                    f"Data directory {args.data_dir} is not an existing directory."
                )
                sys.exit(1)
            self.artifacts_dir = args.data_dir

        self.logger.log_preamble(f"Storing artifacts at {self.artifacts_dir}")
        self.logger.log_preamble(
            f'Starting "{sys.argv[0]}" ({version("gallia")}) with [{" ".join(sys.argv)}]'
        )

        try:
            return asyncio.run(self._run(args))
        except KeyboardInterrupt:
            self.logger.log_critical("ctrl+c received. Terminating…")
            return 128 + signal.SIGINT
        finally:
            self.logger.log_info(
                f"The scan results are located at: {self.artifacts_dir}"
            )


class Scanner(GalliaBase):
    """Base class for all scanner tasks"""

    def __init__(self) -> None:
        super().__init__()
        self.transport: Optional[BaseTransport] = None
        self.power_supply: Optional[PowerSupply] = None
        self.dumpcap: Optional[Dumpcap] = None

    def _add_class_parser(self) -> None:
        super()._add_class_parser()

        group = self.parser.add_argument_group("transport mode related arguments")
        group.add_argument(
            "--target",
            metavar="TARGET",
            default=os.environ.get("GALLIA_TARGET"),
            type=TargetURI,
            help="URI that describes the target",
        )

        group = self.parser.add_argument_group("power supply related arguments")
        group.add_argument(
            "--power-supply",
            metavar="URI",
            default=os.environ.get("GALLIA_POWER_SUPPLY"),
            type=PowerSupplyURI,
            help="URI specifying the location of the relevant opennetzteil server",
        )
        group.add_argument(
            "--power-cycle",
            default=os.environ.get("GALLIA_POWER_CYCLE"),
            const=5.0,
            nargs="?",
            type=float,
            help=(
                "trigger a powercycle before starting the scan; "
                "optional argument specifies the sleep time in secs"
            ),
        )
        group.add_argument(
            "--dumpcap",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Enable/Disable creating a pcap file",
        )

    @staticmethod
    async def _setup_transport(target: TargetURI) -> BaseTransport:
        transports = [
            ISOTPTransport,
            RawCANTransport,
            DoIPTransport,
            TCPLineSepTransport,
        ]

        def func(x: EntryPoint) -> type[BaseTransport]:
            t = x.load()
            if not issubclass(t, BaseTransport):
                raise ValueError(f"{type(x)} is not derived from BaseTransport")
            return cast(type[BaseTransport], t)

        transports_eps = map(func, entry_points()["gallia_transports"])
        transports += transports_eps

        for transport in transports:
            if target.scheme == transport.SCHEME:  # type: ignore
                t = transport(target)
                await t.connect(None)
                return t

        raise ValueError(f"no transport for {target}")

    async def setup(self, args: Namespace) -> None:
        if args.power_supply is not None:
            self.power_supply = await PowerSupply.connect(args.power_supply)
            if (time_ := args.power_cycle) is not None:
                await self.power_supply.power_cycle(time_, lambda: asyncio.sleep(2))
        elif args.power_cycle is not None:
            self.parser.error("--power-cycle needs --power-supply")

        # Start dumpcap as the first subprocess; otherwise network
        # traffic might be missing.
        if args.dumpcap:
            self.dumpcap = await Dumpcap.start(args.target, self.artifacts_dir)
            await self.dumpcap.sync()

        self.transport = await self._setup_transport(args.target)

    async def teardown(self, args: Namespace) -> None:
        if self.transport:
            await self.transport.terminate()
        if self.dumpcap:
            await self.dumpcap.stop()


class UDSScanner(Scanner):
    """Base class for all UDS scanner tasks"""

    def __init__(self) -> None:
        super().__init__()
        self.ecu: ECU
        self.tester_present_task: Optional[Task] = None
        self._implicit_logging = True
        self.log_scan_run = True  # TODO: Remove this as soon as find-endpoint is fixed

    def _add_class_parser(self) -> None:
        super()._add_class_parser()

        group = self.parser.add_argument_group("UDS scanner related arguments")

        eps = entry_points()
        choices = (
            [x.name for x in eps["gallia_ecus"]]
            if "gallia_ecus" in eps
            else ["default"]
        )
        group.add_argument(
            "--oem",
            default=os.environ.get("GALLIA_OEM", "default"),
            choices=choices,
            metavar="OEM",
            help="The OEM of the ECU, used to choose a OEM specific ECU implementation",
        )
        group.add_argument(
            "--timeout",
            default=2,
            type=float,
            metavar="SECONDS",
            help="Timeout value to wait for a response from the ECU",
        )
        group.add_argument(
            "--max-retries",
            default=3,
            type=int,
            metavar="INT",
            help="Number of maximum retries while sending UDS requests",
        )
        group.add_argument(
            "--tester-present-interval",
            default=0.5,
            type=float,
            metavar="SECONDS",
            help="Modify the interval of the cyclic tester present packets",
        )
        group.add_argument(
            "--tester-present",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Enable/Disable tester present background worker",
        )
        group.add_argument(
            "--properties",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="Read and store the ECU proporties prior and after scan",
        )
        group.add_argument(
            "--compare-properties",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="Compare properties before and after the scan",
        )

    def load_ecu(self, vendor: str) -> type[ECU]:
        if vendor == "default":
            return ECU

        eps = entry_points()
        if "gallia_ecus" in eps:
            for entry_point in eps["gallia_ecus"]:
                if vendor == entry_point.name:
                    self.logger.log_debug(f"Loading OEM-ECU {entry_point}")
                    return entry_point.load()

        raise ValueError(f"no such OEM: '{vendor}'")

    async def _tester_present_worker(self, interval: int) -> None:
        assert self.transport
        self.logger.log_debug("tester present worker started")
        while True:
            try:
                async with self.transport.mutex:
                    await self.transport.write(bytes([0x3E, 0x80]), tags=["IGNORE"])

                    # Hold the mutex for 10 ms to synchronize this background
                    # worker with the main sender task.
                    await asyncio.sleep(0.01)

                    # The BCP might send us an error. Everything
                    # will break if we do not read it back. Since
                    # this read() call is only intended to flush
                    # errors caused by the previous write(), it is
                    # sane to ignore the error here.
                    try:
                        await self.transport.read(timeout=0.01)
                    except asyncio.TimeoutError:
                        pass
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                self.logger.log_debug("tester present worker terminated")
                break
            except Exception as e:
                self.logger.log_debug(f"tester present got {type(e)}")
                # Wait until the stack recovers, but not for too long…
                await asyncio.sleep(1)

    @property
    def implicit_logging(self) -> bool:
        return self._implicit_logging

    @implicit_logging.setter
    def implicit_logging(self, value: bool) -> None:
        self._implicit_logging = value

        if self.db_handler is not None:
            self._apply_implicit_logging_setting()

    def _apply_implicit_logging_setting(self) -> None:
        if self._implicit_logging:
            self.ecu.db_handler = self.db_handler
        else:
            self.ecu.db_handler = None

    async def setup(self, args: Namespace) -> None:
        await super().setup(args)

        assert self.transport is not None
        self.ecu = self.load_ecu(args.oem)(
            self.transport,
            timeout=args.timeout,
            max_retry=args.max_retries,
            power_supply=self.power_supply,
        )

        if self.db_handler is not None and self.log_scan_run:
            try:
                await self.db_handler.insert_scan_run(str(args.target))
                self._apply_implicit_logging_setting()
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the scan run to the database: {repr(e)}"
                )

        # Handles connecting to the target and waits
        # until it is ready.
        await self.ecu.wait_for_ecu()
        await self.ecu.connect()

        if args.tester_present:
            coroutine = self._tester_present_worker(args.tester_present_interval)
            self.tester_present_task = asyncio.create_task(coroutine)

            # enforce context switch
            # this ensures, that the task is executed at least once
            # if the task is not executed, task.cancel will fail with CancelledError
            await asyncio.sleep(0)

        if args.properties is True:
            path = self.artifacts_dir.joinpath(FileNames.PROPERTIES_PRE.value)
            async with aiofiles.open(path, "w") as file:
                await file.write(json.dumps(await self.ecu.properties(True), indent=4))
                await file.write("\n")

        if self.db_handler is not None and self.log_scan_run:
            try:
                await self.db_handler.insert_scan_run_properties_pre(
                    await self.ecu.properties()
                )
                self._apply_implicit_logging_setting()
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the properties_pre to the database: {repr(e)}"
                )

    async def teardown(self, args: Namespace) -> None:
        if args.properties is True:
            path = self.artifacts_dir.joinpath(FileNames.PROPERTIES_POST.value)
            async with aiofiles.open(path, "w") as file:
                await file.write(json.dumps(await self.ecu.properties(True), indent=4))
                await file.write("\n")

            path_pre = self.artifacts_dir.joinpath(FileNames.PROPERTIES_PRE.value)
            async with aiofiles.open(path_pre, "r") as file:
                prop_pre = json.loads(await file.read())

            if args.compare_properties and await self.ecu.properties(False) != prop_pre:
                self.logger.log_warning("ecu properties differ, please investigate!")

        if self.db_handler is not None and self.log_scan_run:
            try:
                await self.db_handler.complete_scan_run(
                    await self.ecu.properties(False)
                )
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the scan run to the database: {repr(e)}"
                )

        if self.tester_present_task:
            self.tester_present_task.cancel()
            await self.tester_present_task
        # This must be the last one.
        await super().teardown(args)


class DiscoveryScanner(Scanner):
    async def setup(self, args: Namespace) -> None:
        # Setting up mcp and dumpcap is already implemented in the parent class.
        await super().setup(args)

        if self.db_handler is not None:
            try:
                await self.db_handler.insert_discovery_run(args.target.url.scheme)
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the discovery run to the database: {repr(e)}"
                )



import asyncio
import binascii
import os
import sys
import time
from argparse import Namespace
from pathlib import Path
from typing import Optional

import aiofiles

from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class SaDumpSeeds(UDSScanner):
    """This scanner tries to enable ProgrammingSession and dump seeds for 12h."""

    def __init__(self) -> None:
        super().__init__()

        self.implicit_logging = False

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--session",
            metavar="INT",
            type=auto_int,
            default=0x02,
            help="Set diagnostic session to perform test in",
        )
        self.parser.add_argument(
            "--check-session",
            action="store_true",
            default=False,
            help="Check current session with read DID",
        )
        self.parser.add_argument(
            "--level",
            default=0x11,
            metavar="INT",
            type=auto_int,
            help="Set security access level to request seed from",
        )
        self.parser.add_argument(
            "--send-zero-key",
            metavar="BYTE_LENGTH",
            nargs="?",
            const=96,
            default=0,
            type=int,
            help="Attempt to fool brute force protection by pretending to send a key after requesting a seed "
            "(all zero bytes, length can be specified)",
        )
        self.parser.add_argument(
            "--reset",
            nargs="?",
            const=1,
            default=None,
            type=int,
            help="Attempt to fool brute force protection by resetting the ECU after every nth requested seed.",
        )
        self.parser.add_argument(
            "--duration",
            default=12 * 60,
            type=float,
            metavar="FLOAT",
            help="Run script for N minutes; zero or negative for infinite runtime",
        )
        self.parser.add_argument(
            "--data-record",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help="Append an optional data record to each seed request",
        )

    async def request_seed(self, level: int, data: bytes) -> Optional[bytes]:
        resp = await self.ecu.security_access_request_seed(
            level, data, config=UDSRequestConfig(tags=["ANALYZE"])
        )
        if isinstance(resp, NegativeResponse):
            self.logger.log_warning(f"ECU replied with an error: {resp}")
            return None
        return resp.security_seed

    async def send_key(self, level: int, key: bytes) -> bool:
        resp = await self.ecu.security_access_send_key(
            level + 1, key, config=UDSRequestConfig(tags=["ANALYZE"])
        )
        if isinstance(resp, NegativeResponse):
            self.logger.log_debug(f"Key was rejected: {resp}")
            return False
        self.logger.log_summary(
            f'Unlocked SA level 0x{level:02x} with key "{key.hex()}"! resp: {resp}'
        )
        return True

    def log_size(self, path: os.PathLike, time_delta: float) -> None:
        size = os.path.getsize(path) / 1024
        size_unit = "KiB"
        rate = size / time_delta * 3600 if time_delta != 0 else 0
        rate_unit = "KiB"
        if rate > 1024:
            rate = rate / 1024
            rate_unit = "MiB"
        if size > 1024:
            size = size / 1024
            size_unit = "MiB"
        self.logger.log_notice(
            f"Dumping seeds with {rate:.2f}{rate_unit}/h: {size:.2f}{size_unit}"
        )

    async def main(self, args: Namespace) -> None:
        session = args.session
        self.logger.log_info(f"scanning in session: {session:02x}")

        resp = await self.ecu.set_session(session)
        if isinstance(resp, NegativeResponse):
            self.logger.log_critical(f"could not change to session: {resp}")
            return

        i = -1
        seeds_file = Path.joinpath(self.artifacts_dir, "seeds.bin")
        file = await aiofiles.open(seeds_file, "wb", buffering=0)
        duration = args.duration * 60
        start_time = time.time()
        last_seed = b""
        reset = False
        runs_since_last_reset = 0

        while duration <= 0 or (time.time() - start_time) < duration:
            # Only emit logs every 1kB. Otherwise the logs
            # include too much garbage…
            i = (i + 1) % 1000
            if i == 0:
                self.log_size(seeds_file, time.time() - start_time)
            if args.check_session or reset:
                if not await check_and_set_session(self.ecu, args.session):
                    self.logger.log_error(
                        f"ECU persistently lost session {args.session}"
                    )
                    sys.exit(1)

            reset = False

            try:
                seed = await self.request_seed(args.level, args.data_record)
            except asyncio.TimeoutError:
                self.logger.log_error("Timeout while requesting seed")
                continue
            except Exception as e:
                self.logger.log_critical(f"Error while requesting seed: {e}")
                sys.exit(1)

            if seed is None:
                # Errors are already logged in .request_seed()
                continue

            await file.write(seed)
            if last_seed == seed:
                self.logger.log_warning("Received the same seed as before")

            last_seed = seed

            if args.send_zero_key > 0:
                try:
                    if await self.send_key(args.level, bytes(args.send_zero_key)):
                        break
                except asyncio.TimeoutError:
                    self.logger.log_warning("Timeout while sending key")
                    continue
                except Exception as e:
                    self.logger.log_critical(f"Error while sending key: {e}")
                    sys.exit(1)

            runs_since_last_reset += 1

            if runs_since_last_reset == args.reset:
                reset = True
                runs_since_last_reset = 0

                try:
                    self.logger.log_info("Resetting the ECU")
                    await self.ecu.ecu_reset(0x01)
                    self.logger.log_info("Waiting for the ECU to recover…")
                    await self.ecu.wait_for_ecu()
                except asyncio.TimeoutError:
                    self.logger.log_error("ECU did not respond after reset; exiting…")
                    sys.exit(1)
                except ConnectionError:
                    self.logger.log_warning(
                        "Lost connection to the ECU after performing a reset. "
                        "Attempting to reconnect…"
                    )
                    await self.ecu.reconnect()

                # Re-enter session. Checking/logging will be done at the beginning of next iteration
                await self.ecu.set_session(session)

        await file.close()
        self.log_size(seeds_file, time.time() - start_time)
        await self.ecu.leave_session(session)



import re
import ipaddress
from typing import Optional
from urllib.parse import urlparse


def split_host_port(
    hostport: str,
    default_port: Optional[int] = None,
) -> tuple[str, Optional[int]]:
    """Splits a combination of ip address/hostname + port into hostname/ip address
    and port.  The default_port argument can be used to return a port if it is
    absent in the hostport argument."""
    # Special case: If hostport is an ipv6 then the urlparser does some weird
    # things with the colons and tries to parse ports. Catch this case early.
    host = ""
    port = default_port
    try:
        # If hostport is a valid ip address (v4 or v6) there
        # is no port included
        host = str(ipaddress.ip_address(hostport))
    except ValueError:
        pass

    # Only parse of hostport is not a valid ip address.
    if host == "":
        # urlparse() and urlsplit() insists on absolute URLs starting with "//".
        url = urlparse(f"//{hostport}")
        host = url.hostname if url.hostname else url.netloc
        port = url.port if url.port else default_port
    return host, port


def camel_to_snake(s: str) -> str:
    """Convert a CamelCase string to a snake_case string."""
    # https://stackoverflow.com/a/12867228
    return re.sub(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))", r"_\1", s).lower()


def camel_to_dash(s: str) -> str:
    """Convert a CamelCase string to a dash-case string."""
    return camel_to_snake(s).replace("_", "-")



import binascii
import sys
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse, UDSResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class WriteByIdentifier(UDSScanner):
    """A simple scanner to talk to the write by identifier service"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=0x01,
            help="set session perform test in",
        )
        self.parser.add_argument(
            "--data-id",
            type=auto_int,
            required=True,
            help="which data identifier to write to",
        )
        self.parser.add_argument(
            "--data",
            required=True,
            type=binascii.unhexlify,
            help="data to write as hex",
        )

    async def main(self, args: Namespace) -> None:
        try:
            if args.session != 0x01:
                resp: UDSResponse = await self.ecu.set_session(args.session)
                if isinstance(resp, NegativeResponse):
                    self.logger.log_critical(f"could not change to session: {resp}")
                    sys.exit(1)
        except Exception as e:
            self.logger.log_critical(f"fatal error: {e}")
            sys.exit(1)

        resp = await self.ecu.write_data_by_identifier(args.data_id, args.data)
        if isinstance(resp, NegativeResponse):
            self.logger.log_error(f"{resp}")
        else:
            self.logger.log_info("Positive response")



import sys
from argparse import Namespace
from functools import partial

from tabulate import tabulate

from gallia.uds.core.constants import CDTCSSubFuncs, UDSErrorCodes, DSCSubFuncs
from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class DTCScanner(UDSScanner):
    """cmd_dtc"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session",
            default=DSCSubFuncs.DS.value,
            type=auto_int,
            help="Session to perform test in",
        )
        sub_parser = self.parser.add_subparsers(dest="cmd", required=True)
        read_parser = sub_parser.add_parser(
            "read", help="Read the DTCs using the ReadDTCInformation service"
        )
        read_parser.add_argument(
            "--mask",
            type=partial(int, base=16),
            default=0xFF,
            help="The bitmask which is sent to the ECU in order to select the relevant DTCs according to their "
            "error state. By default, all error codes are returned (c.f. ISO 14229-1,D.2).",
        )
        read_parser.add_argument(
            "--show-legend",
            action="store_true",
            help="Show the legend of the bit interpretation according to ISO 14229-1,D.2",
        )
        read_parser.add_argument(
            "--show-failed",
            action="store_true",
            help="Show a summary of the codes which failed",
        )
        read_parser.add_argument(
            "--show-uncompleted",
            action="store_true",
            help="Show a summary of the codes which have not completed",
        )
        clear_parser = sub_parser.add_parser(
            "clear", help="Clear the DTCs using the ClearDiagnosticInformation service"
        )
        clear_parser.add_argument(
            "--group-of-dtc",
            type=int,
            default=0xFFFFFF,
            help="Only clear a particular DTC or the DTCs belonging to the given group. "
            "By default, all error codes are cleared.",
        )
        control_parser = sub_parser.add_parser(
            "control",
            help="Stop or resume the setting of DTCs using the "
            "ControlDTCSetting service",
        )
        control_group = control_parser.add_mutually_exclusive_group(required=True)
        control_group.add_argument(
            "--stop",
            action="store_true",
            help="Stop the setting of DTCs. If already disabled, this has no effect.",
        )
        control_group.add_argument(
            "--resume",
            action="store_true",
            help="Resume the setting of DTCs. If already enabled, this has no effect.",
        )

    async def fetch_error_codes(self, mask: int, split: bool = True) -> dict[int, int]:
        ecu_response = await self.ecu.read_dtc_information_report_dtc_by_status_mask(
            mask
        )
        dtcs = {}

        if isinstance(ecu_response, NegativeResponse):
            if ecu_response.response_code == UDSErrorCodes.responseTooLong:
                self.logger.log_error(
                    f"There are too many codes for (sub)mask {mask}. Consider setting --mask "
                    f"with a parameter that excludes one or more of the corresponding bits."
                )
                if split:
                    self.logger.log_warning(
                        "Trying to fetch the error codes iteratively."
                    )

                    for i in range(8):
                        sub_mask = mask & 2**i

                        if sub_mask > 0:
                            self.logger.log_info(
                                f"Trying to fetch with mask {sub_mask}"
                            )
                            dtcs.update(await self.fetch_error_codes(sub_mask, False))
            else:
                self.logger.log_critical(
                    f"Could not fetch error codes: {ecu_response}; exiting…"
                )
                sys.exit(1)
        else:
            dtcs = ecu_response.dtc_and_status_record

        return dtcs

    async def read(self, args: Namespace) -> None:
        dtcs = await self.fetch_error_codes(args.mask)

        failed_dtcs: list[list[str]] = []
        uncompleted_dtcs: list[list[str]] = []

        for dtc, error_state in dtcs.items():
            raw_output = f"{dtc:06X} {error_state:02X}"
            bit_string = bin(error_state + 0x100)[
                3:
            ]  # Transform error_state into a bit-string with leading zeros
            table_output = [f"{dtc:06X}", f"{error_state:02X}"] + [
                "X" if b == "1" else "" for b in bit_string
            ]

            # if any kind of test failure
            if error_state & 0xAF:
                self.logger.log_warning(raw_output)
                failed_dtcs.append(table_output)
            # if not failed but also not completed yet (i.e. not yet in this cycle or since last clear)
            elif error_state & 0x50:
                self.logger.log_summary(raw_output)
                uncompleted_dtcs.append(table_output)

        if args.show_legend:
            self.logger.log_summary("")
            self.show_bit_legend()

        if args.show_failed:
            self.logger.log_summary("")
            self.logger.log_summary("Failed codes:")
            self.show_summary(failed_dtcs)

        if args.show_uncompleted:
            self.logger.log_summary("")
            self.logger.log_summary("Uncompleted codes:")
            self.show_summary(uncompleted_dtcs)

    def show_bit_legend(self) -> None:
        bit_descriptions = [
            "0 = testFailed: most recent test failed",
            "1 = testFailedThisOperationCycle: failed in current cycle",
            "2 = pendingDTC: failed in current or last cycle",
            "3 = confirmedDTC: failed often enough to be stored in long term memory",
            "4 = testNotCompletedSinceLastClear: not completed since last clear",
            "5 = testFailedSinceLastClear: failed since last clear",
            "6 = testNotCompletedThisOperationCycle: not completed in current cycle",
            "7 = warningIndicatorRequested: existing warning indicators (e.g. lamp, display)",
        ]

        for line in (
            tabulate([[d] for d in bit_descriptions], headers=["bit descriptions"])
        ).splitlines():
            self.logger.log_summary(line)

    def show_summary(self, dtcs: list[list[str]]) -> None:
        dtcs.sort()

        header = [
            "DTC",
            "error state",
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
        ]

        for line in tabulate(dtcs, headers=header, tablefmt="fancy_grid").splitlines():
            self.logger.log_summary(line)

    async def clear(self, args: Namespace) -> None:
        group_of_dtc: int = args.group_of_dtc

        if not 0 <= group_of_dtc <= 0xFFFFFF:
            self.logger.log_error(
                "The parameter group_of_dtc must be in the range 0x000000-0xffffff"
            )

        resp = await self.ecu.clear_diagnostic_information(group_of_dtc)

        if isinstance(resp, NegativeResponse):
            self.logger.log_error(resp)
        else:
            self.logger.log_summary("Success")

    async def control(self, args: Namespace) -> None:
        if args.stop:
            await self.ecu.control_dtc_setting(CDTCSSubFuncs.OFF)
        else:
            await self.ecu.control_dtc_setting(CDTCSSubFuncs.ON)

    async def main(self, args: Namespace) -> None:
        await self.ecu.set_session(args.session)

        if args.cmd == "clear":
            await self.clear(args)
        elif args.cmd == "control":
            await self.control(args)
        elif args.cmd == "read":
            await self.read(args)
        else:
            self.logger.log_critical("Unhandled command")
            sys.exit(1)



import binascii
import sys
from argparse import Namespace
from pathlib import Path

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class WriteMemoryByAddressScanner(UDSScanner):
    """Write memory by address"""

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=0x01,
            help="The session in which the requests are made",
        )
        self.parser.add_argument(
            "address",
            type=auto_int,
            help="The start address to which data should be written",
        )
        data_group = self.parser.add_mutually_exclusive_group(required=True)
        data_group.add_argument(
            "--data", type=binascii.unhexlify, help="The data which should be written"
        )
        data_group.add_argument(
            "--data-file",
            type=Path,
            help="The path to a file with the binary data which should be written",
        )

    async def main(self, args: Namespace) -> None:
        try:
            await check_and_set_session(self.ecu, args.session)
        except Exception as e:
            self.logger.log_critical(
                f"Could not change to session: 0x{args.session:02x}: {e.__class__.__name__} {e}"
            )
            sys.exit(1)

        if args.data is not None:
            data = args.data
        else:
            with args.data_file.open("rb") as file:
                data = file.read()

        resp = await self.ecu.write_memory_by_address(args.address, data)

        if isinstance(resp, NegativeResponse):
            self.logger.log_error(resp)
        else:
            # There is not real data returned, only echos
            self.logger.log_summary("Success")



import asyncio
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, find_sessions


class ReadErrorLog(UDSScanner):
    """read_error_log"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--sessions",
            type=auto_int,
            nargs="*",
            help="set list of sessions to perform test in, or all",
        )
        self.parser.add_argument(
            "--clear-dtc",
            action="store_true",
            help="Clear DTC log",
        )

    async def main(self, args: Namespace) -> None:
        sessions = args.sessions
        if sessions is None or len(sessions) == 0:
            sessions = list(range(1, 0x80))
            sessions = await find_sessions(self.ecu, sessions)
            msg = f'Found {len(sessions)} sessions: {" ".join([hex(i) for i in sessions])}'
            self.logger.log_summary(msg)

        for sess in sessions:
            await self.ecu.set_session(sess)
            resp = await self.ecu.read_dtc()
            if isinstance(resp, NegativeResponse):
                self.logger.log_warning(resp)
            else:
                self.logger.log_summary(resp.dtc_and_status_record)
            await self.ecu.leave_session(sess)

        if args.clear_dtc:
            await self.ecu.clear_dtc()
            await self.ecu.read_dtc()
            self.logger.log_info("Rebooting ECU...")
            await self.ecu.ecu_reset(1)
            await asyncio.sleep(2)
            await self.ecu.read_dtc()



import asyncio
import sys
from argparse import Namespace
from binascii import unhexlify

from gallia.uds.core.utils import uds_memory_parameters
from gallia.uds.core.constants import UDSErrorCodes
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class ScanWriteDataByAddress(UDSScanner):
    """This scanner scans functions with direct access to memory.
    Specifically, these are service 0x3d WriteMemoryByAddress, 0x34 RequestDownload
    and 0x35 RequestUpload, which all share the same packet structure, except for
    0x3d which requires an additional data field.
    """

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--session", type=auto_int, default=0x03, help="set session to perform test"
        )
        self.parser.add_argument(
            "--check-session",
            nargs="?",
            const=1,
            type=int,
            help="Check current session via read DID [for every nth MemoryAddress] and try to recover session",
        )
        self.parser.add_argument(
            "--sid",
            required=True,
            choices=[0x23, 0x3D, 0x34, 0x35],
            type=auto_int,
            help="Choose between 0x23 ReadMemoryByAddress 0x3d WriteMemoryByAddress, "
            "0x34 RequestDownload and 0x35 RequestUpload",
        )
        self.parser.add_argument(
            "--data",
            default="0000000000000000",
            type=unhexlify,
            help="Service 0x3d requires a data payload which can be specified with this flag as a hex string",
        )

    async def main(self, args: Namespace) -> None:
        resp = await self.ecu.set_session(args.session)
        if isinstance(resp, NegativeResponse):
            self.logger.log_critical(f"could not change to session: {resp}")
            sys.exit(1)

        for i in range(5):
            await self.scan_memory_address(args, i)

    async def scan_memory_address(self, args: Namespace, addr_offset: int = 0) -> None:
        sid = args.sid
        data = args.data if sid == 0x3D else None  # Only service 0x3d has a data field
        memory_size = len(data) if data else 0x1000

        for i in range(0x100):
            addr = i << (addr_offset * 8)

            (
                addr_and_length_identifier,
                addr_bytes,
                mem_size_bytes,
            ) = uds_memory_parameters(addr, memory_size)

            pdu = bytes([sid])
            if sid in [0x34, 0x35]:
                # RequestUpload and RequestDownload require a DataFormatIdentifier
                # byte that defines encryption and compression. 00 is neither.
                pdu += bytes([00])
            pdu += bytes([addr_and_length_identifier])
            pdu += addr_bytes + mem_size_bytes
            pdu += data if data else b""

            if args.check_session and i % args.check_session == 0:
                # Check session and try to recover from wrong session (max 3 times), else skip session
                if not await check_and_set_session(self.ecu, args.session):
                    self.logger.log_error(
                        f"Aborting scan on session 0x{args.session:02x}; "
                        + f"current memory address was 0x{addr:0x}"
                    )
                    sys.exit(1)

            try:
                resp = await self.ecu.send_raw(
                    pdu, config=UDSRequestConfig(tags=["ANALYZE"])
                )
            except asyncio.TimeoutError:
                self.logger.log_summary(f"Address 0x{addr_bytes.hex()}: timeout")
                continue

            if isinstance(resp, NegativeResponse):
                if resp.response_code is UDSErrorCodes.requestOutOfRange:
                    self.logger.log_info(f"Address 0x{addr_bytes.hex()}: {resp}")
                else:
                    self.logger.log_summary(f"Address 0x{addr_bytes.hex()}: {resp}")
            else:
                self.logger.log_summary(f"Address 0x{addr_bytes.hex()}: {resp}")



import asyncio
import sys
from argparse import Namespace
from typing import Union

from gallia.uds.core.constants import UDSErrorCodes
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.service import (
    NegativeResponse,
    DiagnosticSessionControlResponse,
    UDSResponse,
)
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class IterateSessions(UDSScanner):
    """Iterate Sessions"""

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--depth", type=auto_int, default=None, help="Specify max scanning depth."
        )
        self.parser.add_argument(
            "--sleep",
            metavar="SECONDS",
            type=auto_int,
            default=0,
            help="Sleep this amount of seconds after changing to DefaultSession",
        )
        self.parser.add_argument(
            "--skip",
            metavar="SESSION_ID",
            type=auto_int,
            default=[],
            nargs="*",
            help="List with session IDs to skip while scanning",
        )
        self.parser.add_argument(
            "--with-hooks",
            action="store_true",
            help="Use hooks in case of a ConditionsNotCorrect error",
        )
        self.parser.add_argument(
            "--reset", action="store_true", help="Reset the ECU after each iteration"
        )

    async def set_session_with_hooks_handling(
        self, session: int, use_hooks: bool
    ) -> Union[NegativeResponse, DiagnosticSessionControlResponse]:
        resp = await self.ecu.set_session(
            session, config=UDSRequestConfig(skip_hooks=True)
        )

        if (
            isinstance(resp, NegativeResponse)
            and resp.response_code == UDSErrorCodes.conditionsNotCorrect
        ):
            if not use_hooks:
                self.logger.log_warning(
                    f"Session {session:02x} is potentially available but could not be entered. "
                    f"Use --with-hooks to try to enter the session using hooks to scan for "
                    f"transitions available from that session."
                )
                return resp

            resp_ = await self.ecu.set_session(
                session, config=UDSRequestConfig(skip_hooks=False)
            )

            if isinstance(resp, NegativeResponse):
                self.logger.log_notice(
                    f"Received conditionsNotCorrect for session {session:02x}. "
                    f"Successfully changed to the session with hooks."
                )
                resp = resp_

        return resp

    async def recover_stack(self, stack: list, use_hooks: bool) -> bool:
        for session in stack:
            try:
                resp = await self.set_session_with_hooks_handling(session, use_hooks)

                if isinstance(resp, NegativeResponse):
                    self.logger.log_error(
                        f"Could not change to session 0x{session:02x} as part of stack: {resp}. "
                        f"Try with --reset to reset between each iteration."
                    )
                    return False
            except Exception as e:
                self.logger.log_error(
                    f"Could not change to session 0x{session:02x} as part of stack: {e.__class__.__name__}. "
                    f"Try with --reset to reset between each iteration."
                )
                return False
        return True

    async def main(self, args: Namespace) -> None:
        found: dict[int, list[list[int]]] = {0: [[0x01]]}
        positive_results: list[dict] = []
        negative_results: list[dict] = []
        activated_sessions: set[int] = set()

        sessions = list(range(1, 0x80))
        depth = 0

        # pylint: disable=too-many-nested-blocks
        while (args.depth is None or depth < args.depth) and len(found[depth]) > 0:
            depth += 1

            found[depth] = []
            self.logger.log_summary(f"Depth: {depth}")

            for stack in found[depth - 1]:
                if stack:
                    self.logger.log_summary(f"Starting from session: 0x{stack[-1]:02x}")

                for session in sessions:
                    if session in args.skip:
                        self.logger.log_info(
                            f"Skipping session 0x{session:02x} as requested"
                        )
                        continue

                    if args.reset:
                        try:
                            self.logger.log_info("Resetting the ECU")
                            resp: UDSResponse = await self.ecu.ecu_reset(0x01)

                            if isinstance(resp, NegativeResponse):
                                self.logger.log_warning(
                                    f"Could not reset ECU: {resp}; "
                                    f"continue without reset"
                                )
                            else:
                                self.logger.log_info("Waiting for the ECU to recover…")
                                await self.ecu.wait_for_ecu()
                        except asyncio.TimeoutError:
                            self.logger.log_error(
                                "ECU did not respond after reset; exiting…"
                            )
                            sys.exit(1)
                        except ConnectionError:
                            self.logger.log_warning(
                                "Lost connection to the ECU after performing a reset. "
                                "Attempting to reconnect…"
                            )
                            await self.ecu.reconnect()

                    try:
                        resp = await self.ecu.set_session(0x01)
                        if isinstance(resp, NegativeResponse):
                            self.logger.log_error(
                                f"Could not change to default session: {resp}"
                            )
                            sys.exit(1)
                    except Exception as e:
                        self.logger.log_error(
                            f"Could not change to default session: {e.__class__.__name__}"
                        )
                        sys.exit(1)

                    self.logger.log_debug(
                        f"Sleeping for {args.sleep}s after changing to DefaultSession"
                    )
                    await asyncio.sleep(args.sleep)

                    if not await self.recover_stack(stack, args.with_hooks):
                        sys.exit(1)

                    try:
                        resp = await self.set_session_with_hooks_handling(
                            session, args.with_hooks
                        )

                        # do not ignore NCR subFunctionNotSupportedInActiveSession in this case
                        if (
                            isinstance(resp, NegativeResponse)
                            and resp.response_code
                            == UDSErrorCodes.subFunctionNotSupported
                        ):
                            self.logger.log_info(
                                f"Could not change to session 0x{session:02x}: {resp}"
                            )
                            continue

                        self.logger.log_summary(
                            f"Found session: 0x{session:02x} via stack: {stack}; {resp}"
                        )

                        if not isinstance(resp, NegativeResponse):
                            # Reaching each session once is enough
                            if session not in stack:
                                found[depth].append(stack + [session])

                            activated_sessions.add(session)
                            positive_results.append(
                                {"session": session, "stack": stack, "error": None}
                            )
                        else:
                            negative_results.append(
                                {
                                    "session": session,
                                    "stack": stack,
                                    "error": resp.response_code,
                                }
                            )

                    except asyncio.TimeoutError:
                        self.logger.log_warning(
                            f"Could not change to session 0x{session:02x}: Timeout"
                        )
                        continue

        self.logger.log_summary("Scan finished; Found the following sessions:")
        previous_session = 0

        for res in sorted(positive_results, key=lambda x: x["session"]):
            session = res["session"]

            if session != previous_session:
                previous_session = session
                self.logger.log_summary(f"* Session {session:02x} ")

            self.logger.log_summary(
                f"\tvia stack: {'->'.join([f'{i:02x}' for i in res['stack']])}"
            )

        self.logger.log_summary(
            "The following sessions were identified but could not be activated:"
        )
        previous_session = 0

        for res in sorted(negative_results, key=lambda x: x["session"]):
            session = res["session"]

            if (
                session not in activated_sessions
                and res["error"] != UDSErrorCodes.subFunctionNotSupportedInActiveSession
            ):
                if session != previous_session:
                    previous_session = session
                    self.logger.log_summary(f"* Session {session:02x} ")

                self.logger.log_summary(
                    f"\tvia stack: {'->'.join([f'{i:02x}' for i in res['stack']])} "
                    f"(NRC: {res['error']})"
                )



from ctypes import byref, c_char_p, c_uint, c_int, CDLL, create_string_buffer, POINTER
from ctypes.util import find_library
from typing import Callable, Union


DEFAULT_BUFFER_LIMIT = 2**28
UCL_UNAVAILABLE = "libucl is not available!"


libucl = None
if (lib := find_library("ucl")) is not None:
    libucl = CDLL(lib)


# TODO: Loop for large buffers.
def nrv2b_decompress_8(in_data: Union[bytes, bytearray], out_size: int) -> bytes:
    if libucl is None:
        raise RuntimeError(UCL_UNAVAILABLE)
    in_data_ = bytes(in_data)
    f = libucl.ucl_nrv2b_decompress_safe_8
    f.argtypes = [c_char_p, c_uint, c_char_p, POINTER(c_uint), c_char_p]
    f.restype = c_int
    in_buf = create_string_buffer(in_data_)
    out_buf = create_string_buffer(out_size)
    new_len = c_uint(out_size)
    r = f(in_buf, len(in_data_), out_buf, byref(new_len), None)
    if r != 0:
        raise RuntimeError(f"could not decompress nrv data, exit code {r}")
    return bytes(out_buf.raw)


def _execute_nrv_operation_blindly(
    in_data: Union[bytes, bytearray], operation: Callable, buffer_limit: int
) -> bytes:
    """Executes an NRV operation without knowing the size of the result.

    Args:
      in_data: The data to (de)compress.
      operation: A wrapper around) the function which computes the (de)compression function which
                     takes the four parameters (in_buf, in_buffer_size, out_buf, new_len) and
                     returns the return code of the function.
      buffer_limit: The maximum number of bytes which can be allocated for the output buffer.

    Returns:
      The (de)compressed data.

    """
    in_data_ = bytes(in_data)
    in_buffer_size = len(in_data_)
    in_buf = create_string_buffer(in_data_)

    # We will simply try bigger and bigger buffer sizes until it works but let's start with the
    # size of the incoming data
    out_buffer_size = min(buffer_limit, in_buffer_size * 2)
    out_buf = create_string_buffer(out_buffer_size)
    # For some reason this has to be set to something bigger than the new size in case of
    # decompressing
    new_len = c_uint(out_buffer_size)

    while True:
        r = operation(in_buf, in_buffer_size, out_buf, byref(new_len))

        # If the output buffer is too small
        if r == -202:
            out_buffer_size *= 2
            new_len = c_uint(out_buffer_size)

            if out_buffer_size > buffer_limit:
                raise RuntimeError("Trying to consume too much memory!")

            out_buf = create_string_buffer(out_buffer_size)
            continue

        if r != 0:
            raise RuntimeError(f"Error {r}: could not (de)compress nrv data")

        return bytes(out_buf.raw)[: new_len.value]


def nrv2b_decompress_8_blindly(
    in_data: Union[bytes, bytearray], buffer_limit: int = DEFAULT_BUFFER_LIMIT
) -> bytes:
    """Decompress NRV data without the need to know the size of the output.

    Args:
      in_data: The data to be decompressed
      buffer_limit: The maximum number of bytes which can be allocated for the output buffer.

    Returns:
      The decompressed data.

    """
    if libucl is None:
        raise RuntimeError(UCL_UNAVAILABLE)

    f = libucl.ucl_nrv2b_decompress_safe_8
    f.argtypes = [c_char_p, c_uint, c_char_p, POINTER(c_uint), c_char_p]
    f.restype = c_int

    # Wrapper around f to enable code sharing with nrv2b_compress_blindly using
    # _execute_nrv_operation_blindly
    def f_wrapper(a, b, c, d):  # type: ignore
        return f(a, b, c, d, None)

    return _execute_nrv_operation_blindly(in_data, f_wrapper, buffer_limit)


def nrv2b_compress_blindly(
    in_data: Union[bytes, bytearray],
    level: int = 1,
    buffer_limit: int = DEFAULT_BUFFER_LIMIT,
) -> bytes:
    """Compress NRV data without the need to know the size of the output.

    Args:
      in_data: The data to be compressed
      level: The compression level, where 1 is the lowest and 10 is the highest compression.
      buffer_limit: The maximum number of bytes which can be allocated for the output buffer.

    Returns:
      The compressed data.

    """
    if libucl is None:
        raise RuntimeError(UCL_UNAVAILABLE)

    if level < 1 or level > 10:
        raise ValueError("The compression level must be between 1 and 10!")

    f = libucl.ucl_nrv2b_99_compress
    f.argtypes = [
        c_char_p,
        c_uint,
        c_char_p,
        POINTER(c_uint),
        POINTER(c_int),
        c_int,
        POINTER(c_int),
        POINTER(c_int),
    ]
    f.restype = c_int

    # Wrapper around f to enable code sharing with nrv2b_decompress_8_blindly using
    # _execute_nrv_operation_blindly
    def f_wrapper(a, b, c, d):  # type: ignore
        return f(a, b, c, d, None, level, None, None)

    return _execute_nrv_operation_blindly(in_data, f_wrapper, buffer_limit)



import asyncio
import reprlib
import sys
from argparse import Namespace
from typing import Any

from gallia.uds.core.exception import UnexpectedNegativeResponse
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.exception import IllegalResponse
from gallia.uds.core.service import NegativeResponse, UDSResponse
from gallia.uds.helpers import suggests_sub_function_not_supported
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import (
    auto_int,
    check_and_set_session,
    find_sessions,
    ParseSkips,
)


class ScanReset(UDSScanner):
    """cmd_scan_reset"""

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--sessions",
            type=auto_int,
            nargs="*",
            help="Set list of sessions to be tested; all if None",
        )
        self.parser.add_argument(
            "--skip",
            nargs="+",
            default={},
            type=str,
            action=ParseSkips,
            help="""
                 The sub functions to be skipped per session.
                 A session specific skip is given by <session_id>:<sub_functions>
                 where <sub_functions> is a comma separated list of single ids or id ranges using a dash.
                 Examples:
                  - 0x01:0xf3
                  - 0x10-0x2f
                  - 0x01:0xf3,0x10-0x2f
                 Multiple session specific skips are separated by space.
                 """,
        )
        self.parser.add_argument(
            "--skip-check-session",
            action="store_true",
            help="skip check current session",
        )

    async def main(self, args: Namespace) -> None:
        l_ok: dict[int, list[int]] = dict()
        l_timeout: dict[int, list[int]] = dict()
        l_error: dict[int, list[Any]] = dict()

        if args.sessions is None:
            self.logger.log_info("No sessions specified, starting with session scan")
            # Only until 0x80 because the eight bit is "SuppressResponse"
            sessions = list(
                s
                for s in range(1, 0x80)
                if s not in args.skip or args.skip[s] is not None
            )
            sessions = await find_sessions(self.ecu, sessions)
            self.logger.log_summary(
                f'Found {len(sessions)} sessions: {" ".join([hex(i) for i in sessions])}'
            )
        else:
            sessions = list(
                s
                for s in args.sessions
                if s not in args.skip or args.skip[s] is not None
            )

        self.logger.log_info(f"testing sessions {sessions}")
        self.logger.log_info(f"skipping identifiers {reprlib.repr(args.skip)}")

        for session in sessions:  # pylint: disable=too-many-nested-blocks
            self.logger.log_notice(f"Switching to session 0x{session:02x}")
            resp: UDSResponse = await self.ecu.set_session(session)
            if isinstance(resp, NegativeResponse):
                self.logger.log_warning(
                    f"Switching to session 0x{session:02x} failed: {resp}"
                )
                continue

            self.logger.log_summary(f"Scanning in session: 0x{session:02x}")
            l_ok[session] = list()
            l_timeout[session] = list()
            l_error[session] = list()

            for sub_func in range(0x01, 0x80):
                if session in args.skip and sub_func in args.skip[session]:
                    self.logger.log_notice(
                        f"skipping subFunc: 0x{sub_func:02x} because of --skip"
                    )
                    continue

                if not args.skip_check_session:
                    # Check session and try to recover from wrong session (max 3 times), else skip session
                    if not await check_and_set_session(self.ecu, session):
                        self.logger.log_error(
                            f"Aborting scan on session 0x{session:02x}; current sub-func was 0x{sub_func:02x}"
                        )
                        break

                try:
                    try:
                        resp = await self.ecu.ecu_reset(
                            sub_func, config=UDSRequestConfig(tags=["ANALYZE"])
                        )
                        if isinstance(resp, NegativeResponse):
                            if suggests_sub_function_not_supported(resp):
                                self.logger.log_info(f"0x{sub_func:02x}: {resp}")
                            else:
                                l_error[session].append({sub_func: resp.response_code})
                                msg = f"0x{sub_func:02x}: with error code: {resp}"
                                self.logger.log_summary(msg)
                            continue
                    except IllegalResponse as e:
                        self.logger.log_warning(f"{repr(e)}")

                    self.logger.log_summary(f"0x{sub_func:02x}: reset level found!")
                    l_ok[session].append(sub_func)
                    self.logger.log_info("Waiting for the ECU to recover…")
                    await self.ecu.wait_for_ecu()

                    self.logger.log_info("Reboot ECU to restore default conditions")
                    resp = await self.ecu.ecu_reset(0x01)
                    if isinstance(resp, NegativeResponse):
                        self.logger.log_warning(
                            f"Could not reboot ECU after testing reset level 0x{sub_func:02x}"
                        )
                    else:
                        await self.ecu.wait_for_ecu()

                except asyncio.TimeoutError:
                    self.logger.log_error(
                        f"ECU did not respond after reset level 0x{sub_func:02x}; exiting…"
                    )
                    sys.exit(1)
                except ConnectionError:
                    msg = f"0x{sub_func:02x}: lost connection to ECU (post), current session: 0x{session:0x}"
                    self.logger.log_warning(msg)
                    await self.ecu.reconnect()
                    continue

                # We reach this code only for positive responses
                if not args.skip_check_session:
                    try:
                        current_session = await self.ecu.read_session()
                        self.logger.log_summary(
                            f"0x{sub_func:02x}: Currently in session 0x{current_session:02x}, should be 0x{session:02x}"
                        )
                    except UnexpectedNegativeResponse as e:
                        self.logger.log_warning(
                            f"Could not read current session: {e.RESPONSE_CODE.name}"
                        )

                self.logger.log_info(f"Setting session 0x{session:02x}")
                await self.ecu.set_session(session)

            await self.ecu.leave_session(session)

        self.logger.log_summary(f"ok: {l_ok}")
        self.logger.log_summary(f"timeout: {l_timeout}")
        self.logger.log_summary(f"with error: {l_error}")



from typing import Optional, Union

from gallia.uds.core import service
from gallia.uds.core.constants import UDSErrorCodes, UDSIsoServices
from gallia.uds.core.exception import (
    UnexpectedNegativeResponse,
    RequestResponseMismatch,
    MalformedResponse,
)


def raise_for_error(
    response: service.UDSResponse, message: Optional[str] = None
) -> None:
    if isinstance(response, service.NegativeResponse):
        if response.trigger_request is None:
            raise ValueError("The response has not been assigned a trigger request")

        raise UnexpectedNegativeResponse.parse_dynamic(
            response.trigger_request, response, message
        )


def as_exception(
    response: service.NegativeResponse, message: Optional[str] = None
) -> UnexpectedNegativeResponse:
    if response.trigger_request is None:
        raise ValueError("The response has not been assigned a trigger request")

    return UnexpectedNegativeResponse.parse_dynamic(
        response.trigger_request, response, message
    )


def raise_for_mismatch(
    request: service.UDSRequest,
    response: service.UDSResponse,
    message: Optional[str] = None,
) -> None:
    if not response.matches(request):
        raise RequestResponseMismatch(request, response, message)


def _suggests_not_supported(
    response: Union[service.UDSResponse, UDSErrorCodes],
    not_supported_codes: list[UDSErrorCodes],
) -> bool:
    if isinstance(response, service.UDSResponse):
        if not isinstance(response, service.NegativeResponse):
            return False

        response_code = response.response_code
    else:
        response_code = response

    return response_code in not_supported_codes


def suggests_service_not_supported(
    response: Union[service.UDSResponse, UDSErrorCodes]
) -> bool:
    return _suggests_not_supported(
        response,
        [
            UDSErrorCodes.serviceNotSupported,
            UDSErrorCodes.serviceNotSupportedInActiveSession,
        ],
    )


def suggests_sub_function_not_supported(
    response: Union[service.UDSResponse, UDSErrorCodes]
) -> bool:
    return _suggests_not_supported(
        response,
        [
            UDSErrorCodes.serviceNotSupported,
            UDSErrorCodes.serviceNotSupportedInActiveSession,
            UDSErrorCodes.subFunctionNotSupported,
            UDSErrorCodes.subFunctionNotSupportedInActiveSession,
        ],
    )


def suggests_identifier_not_supported(
    response: Union[service.UDSResponse, UDSErrorCodes]
) -> bool:
    return _suggests_not_supported(
        response,
        [
            UDSErrorCodes.serviceNotSupported,
            UDSErrorCodes.serviceNotSupportedInActiveSession,
            UDSErrorCodes.subFunctionNotSupported,
            UDSErrorCodes.subFunctionNotSupportedInActiveSession,
            UDSErrorCodes.requestOutOfRange,
        ],
    )


def parse_pdu(pdu: bytes, request: service.UDSRequest) -> service.UDSResponse:
    parsed_request = service.UDSRequest.parse_dynamic(request.pdu)

    try:
        response = service.UDSResponse.parse_dynamic(pdu)
    except Exception as e:
        if pdu[0] == UDSIsoServices.NegativeResponse:
            response = service.RawNegativeResponse(pdu)
        else:
            response = service.RawPositiveResponse(pdu)

        raise MalformedResponse(request, response, str(e)) from e

    if not response.matches(parsed_request):
        raise RequestResponseMismatch(request, response)

    response.trigger_request = request

    return response



from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner


class GetVin(UDSScanner):
    """Request VIN"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

    async def main(self, args: Namespace) -> None:
        resp = await self.ecu.read_vin()
        if isinstance(resp, NegativeResponse):
            self.logger.log_warning(f"ECU said: {resp}")
            return
        self.logger.log_summary(resp.data_record.hex())



from binascii import hexlify
from math import ceil
from typing import Optional, Any

from gallia.uds.core.constants import UDSIsoServices


def from_bytes(x: bytes) -> int:
    return int.from_bytes(x, "big")


def to_bytes(x: int, length: int) -> bytes:
    return x.to_bytes(length, "big")


def check_range(data: int, name: str, min_value: int, max_value: int) -> None:
    if not min_value <= data <= max_value:
        raise ValueError(
            f"The {name} parameter must be between {int_repr(min_value)} and "
            f"{int_repr(max_value)}"
        )


def check_data_identifier(data_identifier: int) -> None:
    if not 0 <= data_identifier <= 0xFFFF:
        raise ValueError(f"Not a valid dataIdentifier: {int_repr(data_identifier)}")


def check_sub_function(sub_function: int) -> None:
    if not 0 <= sub_function <= 0x7F:
        raise ValueError(f"Not a valid subFunction: {int_repr(sub_function)}")


def check_length(
    pdu: bytes, minimal_length: int = 0, maximal_length: Optional[int] = None
) -> None:
    if len(pdu) < 1:
        raise ValueError("The PDU is empty")

    if len(pdu) < minimal_length:
        raise ValueError("The PDU is incomplete")

    if maximal_length is not None and len(pdu) > maximal_length:
        raise ValueError("The PDU contains more data than expected")


def int_repr(n: int, prefix: bool = True) -> str:
    s = f"{n:x}"

    if len(s) % 2 == 1:
        s = f"0{s}"

    if prefix:
        s = f"0x{s}"

    return s


def any_repr(x: Any) -> str:
    if isinstance(x, bool):
        return repr(x)
    if isinstance(x, int):
        return int_repr(x)
    elif isinstance(x, (bytes, bytearray)):
        return bytes_repr(x)
    elif isinstance(x, list):
        return f'[{", ".join(any_repr(y) for y in x)}]'
    else:
        return str(x)


def bytes_repr(b: bytes, prefix: bool = False, max_length: Optional[int] = 20) -> str:
    if len(b) == 0:
        return "''"

    s = hexlify(b).decode()

    if max_length is not None and len(s) > max_length:
        s = s[: max_length - 3] + "..."

    return s


def service_repr(service_id: int) -> str:
    try:
        return UDSIsoServices(service_id).name
    except Exception:
        return f"Unknown service {int_repr(service_id)}"


def shorten(string: str, width: int = 16) -> str:
    """Shorten a given string to a specified number of characters.
    An ellipsis `…` as added if the string is truncated.
    """
    if len(string) <= width:
        return string
    return f"{string[:width - 1]}…"


def uds_memory_parameters(
    memory_address: int, memory_size: int, address_and_length_fmt: Optional[int] = None
) -> tuple[int, bytes, bytes]:
    """Transfers the address and size into bytes and computes the corresponding
    addressAndLengthFormatIdentifier (cf. ISO 14229-1) and returns all three.
    The resulting parameters are used in several memory-related UDSClient services.
    Optionally, a desired addressAndLengthFormatIdentifier can be passed, which
    will then be used to turn the ints to the specified length as opposed to
    the minimal byte length.

    Args:
      memory_address: The memory address.
      memory_size: The memory size.
      address_and_length_fmt: Optionally specifies the byte lengths of the
          address and size. If given, the first returned value will be equal to
          this one.

    Returns:
      The UDSClient memory parameters as a tuple, comprised of
       - the addressAndLengthFormatIdentifier
         which specifies the byte lengths of the address and size.
       - memory address (in bytes).
       - memory size (in bytes).

    Raises:
      OverflowError: If any of the memory_address or memory_size parameters is
          either negative or cannot be represented by at most 15 bytes or the
          number of bytes as specified in address_and_length_fmt if given.
      ValueError: If address_and_length_fmt is given and has a wrong format,
          i.e. it is negative, exceeds 0xff or is zero in any of the two nibbles.
    """
    if memory_address < 0:
        raise OverflowError("The memory address must not be negative")
    if memory_size < 0:
        raise OverflowError("The memory size must not be negative")

    if address_and_length_fmt is not None:
        # If the format is given explicitly, format the parameters accordingly
        addr_len, size_len = address_and_size_length(address_and_length_fmt)
        addr_bytes = memory_address.to_bytes(addr_len, "big")
        size_bytes = memory_size.to_bytes(size_len, "big")
    else:
        # Otherwise transfer the parameters to the UDSClient request format

        # The byte length is always at least one, as also a zero must be encoded.
        addr_length = int(max(1, ceil(memory_address.bit_length() / 8)))
        size_length = int(max(1, ceil(memory_size.bit_length() / 8)))

        if addr_length > 0xF:
            raise OverflowError(
                "The memory address is too big to be encoded in the "
                "addressAndLengthFormatIdentifier"
            )
        if size_length > 0xF:
            raise OverflowError(
                "The memory size is too big to be encoded in the "
                "addressAndLengthFormatIdentifier"
            )

        addr_bytes = memory_address.to_bytes(addr_length, "big")
        size_bytes = memory_size.to_bytes(size_length, "big")
        address_and_length_fmt = (size_length << 4) | addr_length

    return address_and_length_fmt, addr_bytes, size_bytes


def address_and_size_length(address_and_length_fmt: int) -> tuple[int, int]:
    """Computes the memory_address and memory_size parameter's byte length
    based on the corresponding addressAndLengthFormatIdentifier (cf. ISO 14229-1)
    which is used in several memory-related UDSClient services.

    Args:
      address_and_length_fmt: Specifies the byte lengths of the address and size.

    Returns:
      The byte length of the address and size according to the given format as a
      tuple, comprised of:

       - The byte length of the memory address
       - The byte length of the memory size

    Raises:
        ValueError: If address_and_length_fmt has the wrong format,
            i.e. it is negative, exceeds 0xff or is zero in any of the two nibbles.
    """
    if not 0x00 <= address_and_length_fmt <= 0xFF:
        raise ValueError(
            "The addressAndLengthFormatIdentifier must not be negative "
            "nor exceed 0xff"
        )
    if address_and_length_fmt & 0xF0 == 0:
        raise ValueError(
            "The addressAndLengthFormatIdentifier's first nibble must not be 0"
        )
    if address_and_length_fmt & 0x0F == 0:
        raise ValueError(
            "The addressAndLengthFormatIdentifier's second nibble must not be 0"
        )

    addr_length = address_and_length_fmt & 0x0F
    size_length = (address_and_length_fmt & 0xF0) >> 4

    return addr_length, size_length


def sub_function_split(sub_function: int) -> tuple[int, bool]:
    """
    Returns the subFunction without suppress bit and if the bit was set.

    :return: The subFunction without suppress bit and if the bit was set.
    """
    return sub_function % 0x80, sub_function >= 0x80



import sys
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class ReadByIdentifier(UDSScanner):
    """cmd_readByIdentifier"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session", type=auto_int, default=0x01, help="set session perform test in"
        )
        self.parser.add_argument(
            "--data-id", type=auto_int, default=0x1001, help="data Identified to read"
        )

    async def main(self, args: Namespace) -> None:
        try:
            if args.session != 0x01:
                await self.ecu.set_session(args.session)
        except Exception as e:
            self.logger.log_critical(f"fatal error: {e}")
            sys.exit(1)

        resp = await self.ecu.read_data_by_identifier(args.data_id)
        if isinstance(resp, NegativeResponse):
            self.logger.log_error(resp)
        else:
            self.logger.log_info("Positive response:")
            data = resp.data_record
            self.logger.log_info(f"hex: {data.hex()}")
            self.logger.log_info(f"raw: {repr(data)}")



import json
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Optional

import aiosqlite

from gallia.uds.core import service
from gallia.uds.core.utils import bytes_repr as bytes_repr_


def bytes_repr(data: bytes) -> str:
    return bytes_repr_(data, False, None)


schema_version = "1.0"

DB_SCHEMA = f"""
CREATE TABLE IF NOT EXISTS version (
  schema text unique,
  version text
);
CREATE TABLE IF NOT EXISTS ecu (
  id integer primary key,
  name text,
  oem text,
  manufacturer text,
  oem_uid text
);
CREATE TABLE IF NOT EXISTS address (
  id integer primary key,
  url text not null unique,
  ecu int references ecu(id) on update cascade on delete set null
);
CREATE TABLE IF NOT EXISTS run_meta (
  id integer primary key,
  script text not null,
  arguments json not null check(json_valid(arguments)),
  start_time real not null,
  start_timezone text not null,
  end_time real,
  end_timezone text check((end_timezone is null) = (end_time is null)),
  exit_code int,
  path text
);
CREATE TABLE IF NOT EXISTS error_log (
  level int not null,
  time real not null,
  timezone text not null,
  message text not null,
  meta int references run_meta(id) not null
);
CREATE TABLE IF NOT EXISTS discovery_run (
  id integer primary key,
  protocol str not null,
  meta int references run_meta(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS scan_run (
  id integer primary key,
  address int references address(id) on update cascade on delete set null,
  properties_pre json check(properties_pre is null or json_valid(properties_pre)),
  properties_post json check(properties_post is null or json_valid(properties_post)),
  meta int references run_meta(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS discovery_result (
  id integer primary key,
  run int not null references discovery_run(id) on update cascade on delete cascade,
  address int not null references address(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS scan_result (
  id integer primary key,
  run int not null references scan_run(id) on update cascade on delete cascade,
  log_mode text not null check(log_mode in ('implicit', 'explicit', 'emphasized')),
  state json check(state is null or json_valid(state)),
  request_pdu blob not null,
  request_time real not null,
  request_timezone text not null,
  request_data json check(json_valid(request_data)) not null,
  response_pdu blob,
  response_time real,
  response_timezone text check((response_timezone is null) = (response_time is null)),
  response_data json check(response_data is null or json_valid(response_data)),
  exception text
);

INSERT OR IGNORE INTO version VALUES('main', '{schema_version}');
"""


class LogMode(Enum):
    implicit = "implicit"
    explicit = "explicit"
    emphasized = "emphasized"


class DBHandler:
    def __init__(self, database: Path):
        self.path = database
        self.connection: Optional[aiosqlite.Connection] = None
        self.scan_run: Optional[int] = None
        self.discovery_run: Optional[int] = None
        self.meta: Optional[int] = None

    async def connect(self) -> None:
        assert self.connection is None, "Already connected to the database"

        self.connection = await aiosqlite.connect(self.path)
        await self.connection.execute("PRAGMA foreign_keys = 1")

        # Allows to read the database in parallel to a scan without causing delays or even losing data
        # This setting is persistent for the database and leads to the creation of extra files
        # See https://www.sqlite.org/wal.html for further information
        await self.connection.execute("PRAGMA journal_mode = WAL")

        await self.connection.executescript(DB_SCHEMA)
        await self.check_version()

    async def disconnect(self) -> None:
        assert self.connection is not None, "Not connected to the database"

        await self.connection.commit()
        await self.connection.close()
        self.connection = None

    async def check_version(self) -> None:
        assert self.connection is not None, "Not connected to the database"

        query = 'SELECT version FROM version WHERE schema = "main"'
        cursor: aiosqlite.Cursor = await self.connection.execute(query)
        row = await cursor.fetchone()

        if row is None:
            raise ValueError("No schema version is specified in the database!")

        version = row[0]

        # This could be a much more complex version to support some older version as well
        # Additionally one might bind certain features to this version or trigger a database migration, etc
        if version != schema_version:
            raise ValueError(
                f"The version of the database schema is not supported! ({version} != {schema_version})"
            )

    async def insert_run_meta(
        self, script: str, arguments: list, start_time: datetime, path: Path
    ) -> None:
        assert self.connection is not None, "Not connected to the database"

        query = (
            "INSERT INTO run_meta(script, arguments, start_time, start_timezone, path) VALUES "
            "(?, ?, ?, ?, ?)"
        )
        cursor = await self.connection.execute(
            query,
            (
                script,
                json.dumps(arguments),
                start_time.timestamp(),
                start_time.tzname(),
                str(path),
            ),
        )

        self.meta = cursor.lastrowid

        await self.connection.commit()

    async def complete_run_meta(self, end_time: datetime, exit_code: int) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        query = "UPDATE run_meta SET end_time = ?, end_timezone = ?, exit_code = ? WHERE id = ?"
        await self.connection.execute(
            query, (end_time.timestamp(), end_time.tzname(), exit_code, self.meta)
        )
        await self.connection.commit()

    async def insert_scan_run(self, target: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        await self.connection.execute(
            "INSERT OR IGNORE INTO address(url) VALUES(?)", (target,)
        )

        query = (
            "INSERT INTO scan_run(address, meta) VALUES "
            "((SELECT id FROM address WHERE url = ?), ?)"
        )
        cursor = await self.connection.execute(query, (target, self.meta))

        self.scan_run = cursor.lastrowid
        await self.connection.commit()

    async def insert_scan_run_properties_pre(self, properties_pre: dict) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.scan_run is not None, "Scan run not yet created"

        query = "UPDATE scan_run SET properties_pre = ? WHERE id = ?"
        await self.connection.execute(
            query, (json.dumps(properties_pre), self.scan_run)
        )
        await self.connection.commit()

    async def complete_scan_run(self, properties_post: dict) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.scan_run is not None, "Scan run not yet created"

        query = "UPDATE scan_run SET properties_post = ? WHERE id = ?"
        await self.connection.execute(
            query, (json.dumps(properties_post), self.scan_run)
        )
        await self.connection.commit()

    async def insert_discovery_run(self, protocol: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        query = "INSERT INTO discovery_run(protocol, meta) VALUES (?, ?)"
        cursor = await self.connection.execute(query, (protocol, self.meta))
        self.discovery_run = cursor.lastrowid
        await self.connection.commit()

    async def insert_discovery_result(self, target: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.discovery_run is not None, "Discovery run not yet created"

        await self.connection.execute(
            "INSERT OR IGNORE INTO address(url) VALUES(?)", (target,)
        )

        query = "INSERT INTO discovery_result(address, run) VALUES ((SELECT id FROM address WHERE url = ?), ?)"

        await self.connection.execute(query, (target, self.discovery_run))

    async def insert_scan_result(
        self,
        state: dict,
        request: service.UDSRequest,
        response: Optional[service.UDSResponse],
        exception: Optional[Exception],
        send_time: datetime,
        receive_time: Optional[datetime],
        log_mode: LogMode,
    ) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.scan_run is not None, "Scan run not yet created"

        request_attributes: dict[str, Any] = {
            "service_id": request.service_id,
            "data": bytes_repr(request.data),
        }
        response_attributes: dict[str, Any] = {}

        if isinstance(request, service.SubFunctionRequest):
            request_attributes["sub_function"] = request.sub_function

        for attr, value in request.__dict__.items():
            if not attr.startswith("_"):
                request_attributes[attr] = value

                if isinstance(value, (bytes, bytearray)):
                    request_attributes[attr] = bytes_repr(value)
                elif (
                    isinstance(value, list)
                    and len(value) > 0
                    and isinstance(value[0], (bytes, bytearray))
                ):
                    request_attributes[attr] = list(bytes_repr(v) for v in value)

        if response is not None:
            response_attributes = {"service_id": response.service_id}

            if isinstance(response, service.PositiveResponse):
                response_attributes["data"] = bytes_repr(response.data)

            if isinstance(response, service.SubFunctionResponse):
                response_attributes["sub_function"] = response.sub_function

            for attr, value in response.__dict__.items():
                if not attr.startswith("_") and attr not in ["trigger_request"]:
                    response_attributes[attr] = value

                    if isinstance(value, (bytes, bytearray)):
                        response_attributes[attr] = bytes_repr(value)
                    elif (
                        isinstance(value, list)
                        and len(value) > 0
                        and isinstance(value[0], (bytes, bytearray))
                    ):
                        response_attributes[attr] = list(bytes_repr(v) for v in value)

        query = (
            "INSERT INTO scan_result(run, state, request_pdu, request_time, request_timezone, request_data, "
            "response_pdu, response_time, response_timezone, response_data, exception, log_mode) "
            "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
        )

        await self.connection.execute(
            query,
            (
                self.scan_run,
                json.dumps(state),
                bytes_repr(request.pdu),
                send_time.timestamp(),
                send_time.tzname(),
                json.dumps(request_attributes),
                bytes_repr(response.pdu) if response is not None else None,
                receive_time.timestamp()
                if response is not None and receive_time is not None
                else None,
                receive_time.tzname()
                if response is not None and receive_time is not None
                else None,
                json.dumps(response_attributes) if response is not None else None,
                repr(exception) if exception is not None else None,
                log_mode.name,
            ),
        )

        await self.connection.commit()
